/* 
* Specification and verification of a simple implementation of a (bounded) set with an array.
* Illustrates the usage of ghost variables for data abstraction and better separation of specification and implementation.
* This version uses the {:autocontracts} attribute to take care of frame generation (read/modifies) and class invariant enforcement.
*/

type T = int // for demo purposes (cannot use Set<T> because new T[] is not supported)
class {:autocontracts} Set
{
  // "Public" variables used for specification purposes only  
  ghost var capacity: nat; 
  ghost var elems: set<T>;
 

  // "Private" variables used for implementation purposes
  var list: array<T>;
  var used : nat;

  // Class invariant (must have this nome with the autocontracts attribute)  
  ghost predicate Valid()
  {
    // constraints on implementation variables + abstraction/refinment relation with specification variables
    used <= list.Length && capacity == list.Length && used == |elems| && multiset(elems) == multiset(list[..used])
    && forall i, j :: 0 <= i < j < used ==> list[i] != list[j] // no dups
  }


  // Constructor, initializes the set with a given capacity (> 0)
  constructor (capacity: nat)
    requires capacity > 0
    ensures elems == {}
  {
    // initialize concrete variables
    list := new T[capacity]; 
    used := 0;
    // initialize ghost variables
    elems := {};
    this.capacity := capacity;
  }

  function  size(): nat
    ensures size() == |elems|  
  {
    used
  }  

  method insert(x : T)
    requires x !in elems && size() < capacity
    ensures elems == old(elems) + {x}
  {
    setProp(elems, x);
	  list[used] := x;
    used := used + 1;
    elems := elems + {x};
  }

  method delete(x: T)
    requires x in elems
    ensures elems == old(elems) - {x} 
  {
    setProp(elems, x);
    var i :| 0 <= i < used && list[i] == x;
    list[i] := list[used-1];
    used := used-1;
    elems := elems - {x};
  }

  function contains(x: T) : bool
    ensures contains(x) <==> x in elems  
  {
    setProp(elems, x);
    exists i :: 0 <= i < used && list[i] == x
  }

  lemma setProp(s: set<T>, x: T)
    ensures x in elems <==> x in multiset(elems)
  {
  }
}


// A simple test case to check on the console.
method {:verify false} Main()
{
  var s := new Set(4);
  s.insert(2);
  s.insert(5);
  s.insert(1);
  s.insert(4);
  print "set contents after inserting 2, 5, 1, 4: ", s.list[..s.used], "\n";
  s.delete(2);
  s.delete(1);
  print "set contents after deleting 2, 1: ", s.list[..s.used], "\n";
}
