/* 
* Specification, implementation and verification of a Set implemented with an array.
* Illustrates the usage of ghost variables and abstraction/refinment relations to better separate
* specification (interface) and implementation.
*/

type T = int // for demo purposes, but could be other type 

class {:autocontracts} Set
{
  // (Public) ghost variables used for specification purposes only  
  ghost var elems: set<T>;
 
  // (Private) concrete variables (and constant) with internal representation.
  static const initialCapacity := 10;
  var list: array<T>;
  var used : nat;
  
  // (Private) predicate that formalizes the class invariant.  
 ghost  predicate Valid()
  {
    // constraints on concrete representation
    used <= list.Length && list.Length >= initialCapacity
    && (forall i, j :: 0 <= i < j < used ==> list[i] != list[j]) // no duplicates
    // abstraction relation
    && |elems| == used && elems == (set x | x in multiset(list[..used]))
  }

  // (Public) Constructor, initializes this set as empty.
  constructor ()
    ensures elems == {}
  {
    // initialize concrete variables
    list := new T[initialCapacity]; 
    used := 0;
    // initialize ghost variable
    elems := {};
  }

  // (Public) Obtains the size of the set
  function  size(): nat
    ensures size() == |elems|  
  {
    used
  }  

  // (Public) Inserts a new element x into this set.
  method insert(x : T)
    requires x !in elems
    ensures elems == old(elems) + {x}
  {
    // update concrete variables
    if used == list.Length 
    {
        var oldList := list;
        list := new T[2 * used];
        forall i | 0 <= i < oldList.Length
        {
          list[i] := oldList[i];
        }
        assert list[..used] == oldList[..used];
    }
	  list[used] := x;
    used := used + 1;
    // update ghost variable
    elems := elems + {x};
  }

  // (Public) Deletes an existing element x from this set.
  method delete(x: T)
    requires x in elems
    ensures elems == old(elems) - {x} 
  {
    // update concrete variables
    var i :| 0 <= i < used && list[i] == x;
    list[i] := list[used-1];
    used := used-1;
    // update ghost variable
    elems := elems - {x};
  }

  // (Public) Checks if this set contains an element x.
  function  contains(x: T) : bool
    ensures contains(x) <==> x in elems  
  {
    // query concrete variables
    exists i :: 0 <= i < used && list[i] == x
  }
}

// A simple test scenario checked statically.
method testSet()
{
  var s := new Set();
  s.insert(2);
  s.insert(5);
  assert s.size() == 2;
  assert s.contains(2);
  assert s.contains(5);
  s.delete(2);
  assert s.size() == 1;
  assert s.contains(5);
  assert !s.contains(2);
  s.delete(5);
  assert !s.contains(5);
  assert s.size() == 0;
}

// Examples of test cases with invalid inputs
method testInvalidDelete()
{
  var s := new Set();
  s.insert(1);
  s.delete(2);
}

method testInvalidInsert()
{
  var s := new Set();
  s.insert(1);
  s.insert(1);
}
